<?php

namespace Drupal\pagedesigner_debug\Service;

use Drupal\Core\Database\Connection;
use Drupal\pagedesigner\Entity\Element;
use Drupal\node\Entity\Node;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\ui_patterns\UiPatternsManager;

/**
 * Class for structure debugging.
 */
class StructureDebugger {

  /**
   * The default checks to perform.
   */
  public const ALL_CHECKS = [
    'entity', 'parent', 'container', 'undelete',
    'publish', 'language', 'child', 'delete', 'overflow_child',
    'container_reference', 'single_container',
    'create_container', 'create_container_translation',
  ];

  /**
   * The checks and corrections to be performed.
   *
   * @var string[]
   */
  protected $checks = [];

  /**
   * The strategy to use.
   *
   * @var string
   */
  protected $strategy = 'preserve';

  /**
   * The database object.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection = NULL;

  /**
   * The list of corrections.
   *
   * @var array
   */
  protected $corrections = [];

  /**
   * The list of corrections.
   *
   * @var array
   */
  protected $encountered = [];

  /**
   * The list of corrections.
   *
   * @var array
   */
  protected $errors = [];

  /**
   * The currently processed node.
   *
   * @var \Drupal\node\Entity\Node
   */
  protected $node = null;

  /**
   * The currently processed container.
   *
   * @var \Drupal\pagedesigner\Entity\Element
   */
  protected $container = null;

  /**
   * The node storage manager.
   *
   * @var \Drupal\node\Entity\NodeStorage
   */
  protected $nodeStorageManager = NULL;

  /**
   * The pattern manager.
   *
   * @var \Drupal\ui_patterns\UiPatternsManager
   */
  protected $patternManager = NULL;

  /**
   * The pattern definitions.
   *
   * @var \Drupal\ui_patterns\Definition\PatternDefinition[]
   */
  protected $patternDefinitions = [];

  /**
   * The container data.
   *
   * @var array
   */
  protected $containerIds = [];

  /**
   * Contains the element data keyed by element id.
   *
   * @var array
   */
  protected $elementData = [];

  /**
   * Contains the element children keyed by element id.
   *
   * @var array
   */
  protected $elementChildren = [];

  /**
   * Contains the element parents keyed by element id.
   *
   * @var array
   */
  protected $elementChildrenParents = [];

  /**
   * Contains the element styles keyed by element id.
   *
   * @var array
   */
  protected $elementStyles = [];

  /**
   * Contains the element style parents keyed by element id.
   *
   * @var array
   */
  protected $elementStylesParent = [];

  /**
   * The field content elements.
   *
   * @var array
   */
  protected $fieldcontent = [];

  /**
   * The patterns for each element.
   *
   * @var array
   */
  protected $fieldpattern = [];

  /**
   * Create a new structure debugger.
   *
   * @param \Drupal\Core\Database\Connection $connection
   *   The database connection.
   * @param \Drupal\Core\Entity\EntityTypeManager $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\ui_patterns\UiPatternsManager $pattern_manager
   *   The pattern manager.
   */
  public function __construct( 
    Connection $connection,
    EntityTypeManager $entity_type_manager,
    UiPatternsManager $pattern_manager
  ) {
    $this->connection = $connection;
    $this->nodeStorageManager = $entity_type_manager->getStorage('node');
    $this->patternManager = $pattern_manager;
    $this->patternDefinitions = $this->patternManager->getDefinitions();
    $this->setChecks();
    $this->setStrategy();
  }

  /**
   * Overwrite the default check selection.
   *
   * @param string[] $checks
   *   The checks to perform.
   */
  public function setChecks(array $checks = self::ALL_CHECKS) {
    $this->checks = $checks;
  }

  /**
   * Set the strategy to use.
   *
   * @param string $strategy
   *   The strategy to employ.
   */
  public function setStrategy(string $strategy = 'preserve') {
    $this->strategy = $strategy;
  }

  /**
   * Check the pagedesigner structure of the given nodes and collect errors.
   *
   * @param array $nids
   *   The nodes to check.
   */
  public function checkStructure(array $nids) {
    if (empty($nids)) {
      return;
    }
    
    // Reset data.
    $this->containerIds = [];
    $this->elementData = [];
    $this->elementChildren = [];
    $this->elementChildrenParent = [];
    $this->elementStyles = [];
    $this->elementStylesParent = [];
    $this->fieldcontent = [];
    $this->corrections = [];
    $this->errors = [];

    $defs = \Drupal::service('entity_field.manager')
      ->getFieldStorageDefinitions('node');
    $mappings = $this->nodeStorageManager->getTableMapping();
    /** @var \Drupal\field\Entity\FieldConfigStorage $field  */
    foreach ($defs as $field) {
      if ($field->getType() == 'pagedesigner_item') {

        // Collect all container relationships on this field.
        $table = $mappings->getFieldTableName($field->getName());
        $columns = $mappings->getColumnNames($field->getName());
        $select = $this->connection->select('node_field_data', 'nfd');
        $select->addField('nfd', 'default_langcode');
        $select->addJoin('INNER', $table, 'nfpc', 'nfd.nid = nfpc.entity_id AND nfd.langcode = nfpc.langcode');
        $select->addField('nfpc', 'entity_id');
        $select->addField('nfpc', $columns['target_id']);
        $select->addField('nfpc', 'langcode');
        $select->condition('nfd.nid', $nids, 'IN');
        $select->orderBy('nfpc.entity_id');
        $select->orderBy('nfd.default_langcode', 'DESC');
        $select->orderBy($columns['target_id'], 'DESC');
        $data = $select->execute();
        $defaultLangcode = "";
        foreach ($data as $row) {
          if ($row->default_langcode == 0) {
            $this->addCorrection(
              $this->containerIds[$row->entity_id],
              'single_container',
              [$row->entity_id, $defaultLangcode]
            );
          }
          else {
            $defaultLangcode = $row->langcode;
            if (in_array($row->{$columns['target_id']}, $this->containerIds)) {
              $nodes = array_keys($this->containerIds, $row->{$columns['target_id']});
              $nodes = array_unique($nodes);
              $nodes[] = $row->entity_id;
              $this->addCorrection($row->{$columns['target_id']}, 'container_reference', $nodes);
            }
            else {
              $this->containerIds[$row->entity_id] = $row->{$columns['target_id']};
            }
          }
        }
      }
    }

    // Collect all element data.
    $data = $this->connection->query('SELECT * FROM pagedesigner_element_field_data')->fetchAll();
    foreach ($data as $row) {
      $this->elementData[$this->makeKey($row->langcode, $row->id)] = $row;
    }

    // Collect children relationship.
    $data = $this->connection->query('SELECT * FROM pagedesigner_element__children')->fetchAll();
    foreach ($data as $row) {
      // Collect all children keyed by parent.
      if (empty($this->elementChildren[$this->makeKey($row->langcode, $row->entity_id)])) {
        $this->elementChildren[$this->makeKey($row->langcode, $row->entity_id)] = [];
      }
      $this->elementChildren[$this->makeKey($row->langcode, $row->entity_id)][] = $row;

      // Collect all parents keyed by child.
      if (empty($this->elementChildrenParent[$this->makeKey($row->langcode, $row->children_target_id)])) {
        $this->elementChildrenParent[$this->makeKey($row->langcode, $row->children_target_id)] = [];
      }
      $this->elementChildrenParent[$this->makeKey($row->langcode, $row->children_target_id)][] = $row;
    }

    // Collect style relationships.
    $data = $this->connection->query('SELECT * FROM pagedesigner_element__field_styles')->fetchAll();
    foreach ($data as $row) {

      // Collect all children keyed by parent.
      if (empty($this->elementStyles[$this->makeKey($row->langcode, $row->entity_id)])) {
        $this->elementStyles[$this->makeKey($row->langcode, $row->entity_id)] = [];
      }
      $this->elementStyles[$this->makeKey($row->langcode, $row->entity_id)][] = $row;

      // Collect all parents keyed by child.
      if (empty($this->elementStylesParent[$this->makeKey($row->langcode, $row->field_styles_target_id)])) {
        $this->elementStylesParent[$this->makeKey($row->langcode, $row->field_styles_target_id)] = [];
      }
      $this->elementStylesParent[$this->makeKey($row->langcode, $row->field_styles_target_id)][] = $row;
    }

    // Collect all content elements.
    $data = $this->connection->query('SELECT * FROM pagedesigner_element__field_content')->fetchAll();
    foreach ($data as $row) {
      $this->fieldcontent[$this->makeKey($row->langcode, $row->entity_id)][] = $row;
    }

    // Collect all patterns.
    $data = $this->connection->query('SELECT * FROM pagedesigner_element__field_pattern')->fetchAll();
    foreach ($data as $row) {
      $this->fieldpattern[$this->makeKey($row->langcode, $row->entity_id)] = $row;
    }

    // Loop through the given nodes.
    foreach ($nids as $nid) {
      $this->node = $this->nodeStorageManager->load($nid);
      $containerId = $this->containerIds[$this->node->id()];
      $this->container = $this->elementData[$this->makeKey($this->node->langcode->value, $containerId)];
      if (empty($this->container)) {
        $this->addCorrection($this->node->id() * -1, 'create_container', $this->node->id());
        continue;
      }

      $this->encountered[] = $containerId;
      // Loop through all node languages.
      foreach ($this->node->getTranslationLanguages() as $language) {
        $this->node = $this->node->getTranslation($language->getId());
        $this->container = $this->elementData[$this->makeKey($this->node->langcode->value, $containerId)];
        if (empty($this->container)) {
          $this->addCorrection($this->makeKey($containerId, $this->node->langcode->value), 'create_container_translation', $this->node->langcode->value);
          continue;
        }
        if ($this->container->deleted != 0) {
          // Containers always need to be published and not marked as deleted.
          $this->addCorrection($this->makeKey($containerId, $this->node->langcode->value), 'undelete', $this->node->langcode->value);
        }

        $children = $this->elementChildren[$this->makeKey($language->getId(), $this->container->id)];
        // Recursively check tree.
        if (!empty($children)) {
          foreach ($children as $row) {
            if (!empty($row->children_target_id)) {
              $this->doCheckStructure($row->children_target_id, $this->container);
            }
          }
        }
      }
    }

    // Clear up errors.
    $errors = $this->errors;
    foreach ($errors as $element => $data) {
      foreach ($data as $type => $value) {
        if ($type == 'multiple parents') {
          if (count(array_intersect($value, $this->encountered)) <= 1) {
            unset($this->errors[$element]['multiple parents']);
            if (empty($this->errors[$element])) {
              unset($this->errors[$element]);
            }
          }
        }
      }
    }
  }

  /**
   * Recursively run through a pagedesigner tree and collect corrections.
   *
   * @param int $element
   *   The id of the element to check.
   * @param object $parentInfo
   *   The info object of the parent.
   */
  protected function doCheckStructure(int $element, object $parentInfo) {
    $this->encountered[] = $element;
    $elementKey = $this->makeKey($parentInfo->langcode, $element);
    $elementData = $this->elementData[$elementKey];
    $pattern = $this->fieldpattern[$elementKey]->field_pattern_value;

    $numFields = 9999999;
    if (!empty($this->patternDefinitions[$pattern])) {
      $numFields = count($this->patternDefinitions[$pattern]->getFields());
    }

    if ($elementData->type == 'style') {
      $parentRows = $this->elementStylesParent[$elementKey] ?: [];
    }
    else {
      $parentRows = ($this->elementChildrenParent[$elementKey]) ?: [];
    }

    // If there are more than two parents.
    if (count($parentRows) > 1) {
      // Collect possible parents.
      // Check whether we have a match on parent field of element.
      $parents = [];
      $parentCorrectionPossible = FALSE;
      $falseParents = [];
      foreach ($parentRows as $row) {
        $parents[] = $row->entity_id;
        if ($row->entity_id == $elementData->parent) {
          $parentCorrectionPossible = TRUE;
        }
        else {
          $falseParents[] = $row->entity_id;
        }
      }
      if ($parentCorrectionPossible) {
        // There is a match on the element parent field.
        $this->addCorrection($element, 'child', $falseParents);
      }
      else {
        $this->errors[$element] = ['multiple parents' => $parents];
      }
    }
    if (empty($elementData->entity)) {
      // If the entity field is empty.
      $this->addCorrection($element, 'entity', $this->node->id());
    }
    elseif ($elementData->entity > $this->node->id() && count($parentRows) == 1) {
      // If we have a mismatch on node and element entity field.
      // We assume content is always copied from older nodes to newer nodes.
      $this->addCorrection($element, 'entity', $this->node->id());
    }
    if ($elementData->langcode != $this->node->langcode->value) {
      // If the language of the element is not matching node language.
      $this->addCorrection($element, 'language', $this->node->langcode->value);
    }
    if ($elementData->container != $this->container->id) {
      // If the language of the element is not matching node language.
      $this->addCorrection($element, 'container', $this->container->id);
    }
    if (count($parentRows) == 1) {
      if ($elementData->parent != $parentInfo->id) {
        // Only one parent in database and parent field does not match.
        $this->addCorrection($element, 'parent', $parentInfo->id);
      }
      else {
        $fieldContent = $this->fieldcontent[$elementKey] ?? [];
        if (
          count($fieldContent) > 0 &&
          !$elementData->status &&
          $parentInfo->status
          ) {
          // Element is content and not published, but parent is.
          $this->addCorrection($element, 'publish', $elementData->langcode);
          if ($elementData->deleted == 1) {
            // Element is also marked as deleted.
            $this->addCorrection($element, 'undelete', $elementData->langcode);
          }
        }
        if ($elementData->deleted == 1) {
          if ($elementData->status) {
            // Undelete published element marked as deleted.
            $this->addCorrection($element, 'undelete', $elementData->langcode);
          }
          else {
            // Remove deleted, unpublished elements from tree.
            $this->addCorrection($element, 'delete', [$elementData->parent]);
          }
        }
      }
    }

    $i = 0;
    if (!empty($this->elementChildren[$elementKey])) {
      foreach ($this->elementChildren[$elementKey] as $row) {
        if (!empty($row->children_target_id)) {
          if ($i >= $numFields) {
            $this->addCorrection($row->children_target_id, 'overflow_child', $element);
          }
          else {
            $this->doCheckStructure($row->children_target_id, $elementData);
            $i++;
          }
        }
      }
    }
    if (!empty($this->elementStyles[$elementKey])) {
      foreach ($this->elementStyles[$elementKey] as $row) {
        if (!empty($row->field_styles_target_id)) {
          $this->doCheckStructure($row->field_styles_target_id, $elementData);
        }
      }
    }
  }

  /**
   * Add a correction.
   *
   * @param string $element
   *   The element id.
   * @param string $type
   *   The type of correction.
   * @param mixed $value
   *   The value for the correction.
   */
  protected function addCorrection(string $element, string $type, $value) {
    if (in_array($type, $this->checks)) {
      if (empty($this->corrections[$element])) {
        $this->corrections[$element] = [];
      }
      $this->corrections[$element][$type] = $value;
    }
  }

  /**
   * Returns the corrections found in the last run.
   *
   * @return array
   *   The corrections of the last run.
   */
  public function getCorrections() {
    return $this->corrections;
  }

  /**
   * Returns the errors found in the last run.
   *
   * @return array
   *   The errors of the last run.
   */
  public function getErrors() {
    return $this->errors;
  }

  /**
   * Set the corrections to apply.
   *
   * @param array $corrections
   *   The corrections to apply.
   */
  public function setCorrections(array $corrections) {
    $this->corrections = $corrections;
  }

  /**
   * Applies the found corrections.
   */
  public function applyCorrections() {

    // Set the entity field to null for special checks.
    $this->nullifyEntity();

    foreach ($this->corrections as $elementId => $corrections) {
      if (strpos($elementId, ':') !== FALSE) {
        $elementId = explode(':', $elementId)[0];
      }
      $element = Element::load($elementId);
      $saveElement = FALSE;
      foreach ($corrections as $type => $value) {
        switch ($type) {
          case 'entity';
            if (in_array($type, $this->checks)) {
              $element->entity->target_id = $value;
              $saveElement = TRUE;
            }
            break;

          case 'container';
            if (in_array($type, $this->checks)) {
              $element->container->target_id = $value;
              $saveElement = TRUE;
            }
            break;

          case 'language';
            if (in_array($type, $this->checks)) {
              $element->langcode->value = $value;
              $saveElement = TRUE;
            }
            break;

          case 'publish';
            if (in_array($type, $this->checks)) {
              if ($element->hasTranslation($value)) {
                $translation = $element->getTranslation($value);
                $translation->status->value = 1;
                $translation->save();
                $saveElement = TRUE;
              }
            }
            break;

          case 'unpublish';
            if (in_array($type, $this->checks)) {
              $element->status->value = 0;
              $saveElement = TRUE;
            }
            break;

          case 'undelete';
            if (in_array($type, $this->checks)) {
              if ($element->hasTranslation($value)) {
                $translation = $element->getTranslation($value);
                $translation->deleted->value = 0;
                $translation->save();
              }
            }
            break;

          case 'delete';
            if (in_array($type, $this->checks)) {
              $element->deleted->value = 1;
              $parent = Element::load($element->parent->target_id);
              $this->removeFromParent($element, $parent);
            }
            break;

          case 'child';
            if (in_array($type, $this->checks)) {
              foreach ($value as $parentId) {
                $parent = Element::load($parentId);
                $this->removeFromParent($element, $parent);
              }
              $element->deleted->value = 0;
              $saveElement = TRUE;
            }
            break;

          case 'parent';
            if (in_array($type, $this->checks)) {
              if ($element->parent->target_id != $value) {
                $element->parent->target_id = $value;
                $saveElement = TRUE;
              }
            }
            break;

          case 'overflow_child';
            if (in_array($type, $this->checks)) {
              $parent = Element::load($value);
              if (!empty($parent)) {
                $this->removeFromParent($element, $parent);
                if (!empty($element->parent->entity)) {
                  $this->removeFromParent($element, $element->parent->entity);
                }
              }
              $element->delete();
            }
            break;

          case 'create_container';
            if (in_array($type, $this->checks)) {
              Node::load($value)->save();
            }
            break;

          case 'create_container_translation';
            if (in_array($type, $this->checks)) {
              if (!$element->hasTranslation($value)) {
                $translation = $element->addTranslation($value);
                $translation->deleted->value = 0;
                $translation->setPublished(FALSE);
                $translation->save();
              }
            }
            break;

          // Delete node translation references to containers.
          case 'single_container';
            if (in_array($type, $this->checks)) {
              $defs = \Drupal::service('entity_field.manager')
                ->getFieldStorageDefinitions('node');
              $mappings = $this->nodeStorageManager->getTableMapping();
              /** @var \Drupal\field\Entity\FieldConfigStorage $field  */
              foreach ($defs as $field) {
                if ($field->getType() == 'pagedesigner_item') {
                  $table = $mappings->getFieldTableName($field->getName());
                  $delete = $this->connection->delete($table);
                  $delete->condition('entity_id', $value[0]);
                  $delete->condition('langcode', $value[1], 'NOT LIKE');
                  $delete->execute();
                }
              }
            }
            break;

          // Make containers unique per node.
          case 'container_reference';
            if (in_array($type, $this->checks)) {
              $source = NULL;
              $nids = $value;
              if (count($nids) > 1) {
                $source = Node::load($nids[0]);
                array_shift($nids);
                foreach ($nids as $nid) {
                  $target = Node::load($nid);
                  \Drupal::service('pagedesigner.service')->overrideContainers($target);
                  \Drupal::service('pagedesigner_duplication.service.duplicator')->duplicate($source, $target);
                  if ($element->isPublished()) {
                    \Drupal::service('pagedesigner.service.statechanger')->publish($target);
                  }
                }
              }
            }
            break;
        }
      }
      if ($saveElement) {
        $element->save();
      }
    }
  }

  /**
   * Removes the element from a given parent element.
   *
   * @param \Drupal\pagedesigner\Entity\Element $element
   *   The element to remove.
   * @param \Drupal\pagedesigner\Entity\Element $parent
   *   The parent to remove the element from.
   */
  protected function removeFromParent(Element $element, Element $parent = NULL) {
    if (empty($parent)) {
      return;
    }
    $parent = $parent->getTranslation($element->langcode->value);
    if ($element->bundle() == 'style') {
      foreach ($parent->field_styles as $delta => $item) {
        if ($item->target_id == $element->id()) {
          $parent->field_styles->removeItem($delta);
        }
      }
    }
    else {
      foreach ($parent->children as $delta => $item) {
        if ($item->target_id == $element->id()) {
          $parent->children->removeItem($delta);
        }
      }
    }
    $parent->save();
  }

  /**
   * Set the entity to null for non-entity elements.
   */
  public function nullifyEntity() {
    $eids = \Drupal::entityQuery('pagedesigner_element')
      ->condition('type', ['layout', 'gallery_item', 'gallery_gallery'], 'IN')
      ->execute();
    foreach ($eids as $eid) {
      $element = Element::load($eid);
      $this->setEntityNull($element);
    }
  }

  /**
   * Set the entity to null for non-entity elements and its children.
   *
   * @param \Drupal\pagedesigner\Entity\Element $element
   *   The element to correct.
   */
  protected function setEntityNull(Element $element) {
    if (!$element->entity->isEmpty()) {
      $element->entity->entity = NULL;
      $element->save();
    }
    foreach ($element->children as $item) {
      if (!empty($item->entity)) {
        $this->setEntityNull($item->entity);
      }
    }
    if ($element->hasField('field_styles')) {
      foreach ($element->field_styles as $item) {
        if (!empty($item->entity)) {
          $this->setEntityNull($item->entity);
        }
      }
    }
  }

  /**
   * Remove orphaned elements.
   *
   * Assumes alls checks on all nodes to ensure no false deletions.
   *
   * @param bool $delete
   *   Whether to mark the orphans as deleted.
   *
   * @return mixed
   *   Returns the number of corrected orphans or FALSE,
   *   if there are corrections or no encountered elements.
   */
  public function removeOrphans(bool $delete = FALSE) {
    if ((empty($this->corrections) && !empty($this->encountered))) {
      // Check for orphaned elements.
      $result = $this->connection->query(
        "SELECT id FROM pagedesigner_element_field_data WHERE `entity` IS NOT NULL AND COALESCE(deleted, 0) = 0");
      $rows = $result->fetchAll();
      foreach ($rows as $row) {
        $candidates[] = $row->id;
      }
      $orphaned = array_diff($candidates, $this->encountered);
      if ($delete) {
        foreach ($orphaned as $element => $eid) {
          $element = Element::load($eid);
          foreach ($element->getTranslationLanguages() as $language) {
            $element = $element->getTranslation($language->getId());
            $element->deleted->value = 1;
            $element->setPublished(FALSE);
            $element->save();
          }
        }
      }
      return count($orphaned);
    }
    return FALSE;
  }

  /**
   * Create a key to store and read element data.
   *
   * @param string $first
   *   First part of the key.
   * @param string $second
   *   Second part of the key.
   *
   * @return string
   *   The key.
   */
  protected function makeKey($first, $second) {
    return $first . ':' . $second;
  }
}
